#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <gtk/gtk.h>
/* #include <gdk/gdkx.h> */
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "fb.h"
#include "config.h"

#ifdef MEMWATCH
# include "memwatch.h"
#endif

#include "images/icon_reload_20x20.xpm"
#include "images/icon_ok_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"

/* File browser structure. */
typedef struct {

	gboolean	initialized;
	gboolean	map_state;
	GtkAccelGroup	*accelgrp;

	GtkWidget	*file_browser,  /* GtkFileSelection. */
			*type_combo,	/* GtkCombo. */
			*refresh_btn,
			*refresh_btn_label,
			*ok_btn,
			*ok_btn_label,
			*cancel_btn,
			*cancel_btn_label;

        GList           *type_combo_glist;

	gchar		*last_dir_path;

} fb_data_struct;


/* Callbacks. */
static void FileBrowserRefreshCB(GtkWidget *widget, gpointer data);
static void FileBrowserOKCB(GtkWidget *widget, gpointer data);
static void FileBrowserCancelCB(GtkWidget *widget, gpointer data);
static gint FileBrowserDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void FileBrowserTypeChangeCB(GtkWidget *widget, gpointer data);
#if 0
static void FileBrowserSelectListItem(GtkWidget *w, const gchar *s);
#endif

/* Front ends. */
gint FileBrowserInit(void);
void FileBrowserSetTransientFor(GtkWidget *w);
gbool FileBrowserIsQuery(void);
void FileBrowserBreakQuery(void);
gbool FileBrowserGetResponse(
	const gchar *title,
	const gchar *ok_label, const gchar *cancel_label,
        const gchar *path,
        fb_type_struct **type, gint total_types,
        gchar ***path_rtn, gint *path_total_rtns,
        fb_type_struct **type_rtn
);
void FileBrowserMap(void);
void FileBrowserUnmap(void);
void FileBrowserShutdown(void);

/* File browser file type extension management. */
gint FileBrowserTypeListNew(
        fb_type_struct ***list, gint *total,
        const gchar *ext,	/* Space separated list of extensions. */
        const gchar *name	/* Descriptive name. */
);
void FileBrowserDeleteTypeList(
        fb_type_struct **t, gint total
);


#define FB_BTN_WIDTH		(100 + (2 * 3))
#define FB_BTN_HEIGHT		(30 + (2 * 3))

#define FB_DEFAULT_TYPE_STR	"All files (*.*)"


static gint block_loop_level;
static fb_data_struct file_browser_data;
static gbool fb_got_user_response;

static gint fb_response_total_paths;
static gchar **fb_response_path;

static fb_type_struct fb_response_type;


/*
 *	Refresh button signal callback.
 */
static void FileBrowserRefreshCB(GtkWidget *widget, gpointer data)
{
	const gchar *cstrptr;
	gchar *s;
	GtkFileSelection *fs;
        fb_data_struct *fb = (fb_data_struct *)data;
        if(fb == NULL)
            return;

        if(!fb->initialized)
            return;

	fs = GTK_FILE_SELECTION(fb->file_browser);
	if(fs == NULL)
	    return;

	cstrptr = gtk_file_selection_get_filename(fs);
	s = (cstrptr != NULL) ? g_strdup(cstrptr) : NULL;
	if(s != NULL)
	    gtk_file_selection_set_filename(fs, s);
	g_free(s);
}

/*
 *	File browser ok button signal callback.
 */
static void FileBrowserOKCB(GtkWidget *widget, gpointer data)
{
	const gchar *cstrptr;
        GtkFileSelection *fs;
	fb_data_struct *fb = (fb_data_struct *)data;
        if(fb == NULL)
            return;

        if(!fb->initialized)
            return;

	/* Get pointer to file selection widget. */
        fs = GTK_FILE_SELECTION(fb->file_browser);

	/* Mark that we got response. */
	fb_got_user_response = TRUE;

	/* Get selected path. */
	cstrptr = (const gchar *)gtk_file_selection_get_filename(fs);
	if(cstrptr != NULL)
	{
	    gint n;

	    /* Append response path. */
	    if(fb_response_total_paths < 0)
		fb_response_total_paths = 0;

	    n = fb_response_total_paths;
	    fb_response_total_paths++;
	    fb_response_path = (gchar **)g_realloc(
		fb_response_path,
		fb_response_total_paths * sizeof(gchar *)
	    );
	    if(fb_response_path == NULL)
	    {
		fb_response_total_paths = 0;
	    }
	    else
	    {
		fb_response_path[n] = g_strdup(cstrptr);
	    }
	}

	/* Unmap. */
	FileBrowserUnmap();

	/* Break out of blocking loop. */
	gtk_main_quit();
	block_loop_level--;
}

/*
 *	File browser cancel button signal callback.
 */
static void FileBrowserCancelCB(GtkWidget *widget, gpointer data)
{
        fb_data_struct *fb = (fb_data_struct *)data;
        if(fb == NULL)
            return;

        if(!fb->initialized)
            return;

        /* Set responses. */
        fb_got_user_response = FALSE;

	/* Unmap. */
        FileBrowserUnmap();

        /* Break out of blocking loop. */
        gtk_main_quit();
	block_loop_level--;
}           

/*
 *      File browser toplevel "delete_event" signal callback.
 */
static gint FileBrowserDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	FileBrowserCancelCB(widget, data);
	return(TRUE);
}


/*
 *	File browser type combo box change callback.
 */
static void FileBrowserTypeChangeCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w, *entry;
	gchar *type_str, *strptr;
	fb_data_struct *fb = (fb_data_struct *)data;
	if(fb == NULL)
	    return;

	if(!fb->initialized)
	    return;

	w = fb->type_combo;
	if(w == NULL)
	    return;

	/* Get value from type combo box. */
	entry = GTK_COMBO(w)->entry;
	type_str = gtk_entry_get_text(GTK_ENTRY(entry));
	if(type_str == NULL)
	    return;


	/* Begin parsing extension type response values, format is:
	 *
	 *	"<name> (<ext>)"
	 *
	 * Where <ext> is a space seperated list of extensions.
	 */

	/* Seek past spaces. */
	while(ISBLANK(*type_str))
	    type_str++;

	/* Name. */
	g_free(fb_response_type.name);
	fb_response_type.name = NULL;

	strptr = strchr(type_str, '(');
	if(strptr != NULL)
	{
	    fb_response_type.name = g_strdup(type_str);

	    if(fb_response_type.name != NULL)
	    {
	        /* Seek to '(' char and deliminate first space. */
	        strptr = strchr(fb_response_type.name, '(');
	        if(strptr != NULL)
	        {
		    while(strptr > fb_response_type.name)
		    {
		        if(ISBLANK(*strptr))
		        {
			    *strptr = '\0';
			    break;
		        }
		        strptr--;
		    }
		}
	    }
	}

	/* Extension. */
        g_free(fb_response_type.ext);
        fb_response_type.ext = NULL;

	strptr = strchr(type_str, '(');
	if(strptr == NULL)
	{
	    /* No name to parse through. */
	    fb_response_type.ext = g_strdup(type_str);
	}
	else
	{
	    strptr++;	/* Seek past left parens. */

	    /* Seek past initial spaces. */
	    while(ISBLANK(*strptr))
		strptr++;

	    fb_response_type.ext = g_strdup(strptr);
	    if(fb_response_type.ext != NULL)
	    {
		strptr = strchr(fb_response_type.ext, ')');
		if(strptr != NULL)
		{
		    (*strptr) = '\0';

		    /* Strip tailing spaces. */
		    strptr--;
		    while(strptr > fb_response_type.ext)
		    {
			if(ISBLANK(*strptr))
			    (*strptr) = '\0';
			else
			    break;
			strptr--;
		    }
		}
	    }
	}
}

#if 0
/*
 *	Selects the item on the GtkList widget w that matches s.
 */
static void FileBrowserSelectListItem(GtkWidget *w, const gchar *s)
{
	if((w == NULL) || (s == NULL))
	    return;

	if(GTK_IS_LIST(w))
	{
	    GtkList *list = GTK_LIST(w);
	    GList *glist = list->children;
	    GtkListItem *item;


	}
}
#endif

/*
 *	Initializes file browser.
 */
gint FileBrowserInit(void)
{
	gint border_major = 5;
	GList *glist;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent;
	GtkFileSelection *fs;
	fb_data_struct *fb = &file_browser_data;


	/* Reset local globals. */
        block_loop_level = 0;
        fb_got_user_response = FALSE;
        fb_response_path = NULL;
	fb_response_total_paths = 0;
	memset(&fb_response_type, 0x00, sizeof(fb_type_struct));


	/* Reset values. */
        fb->initialized = TRUE;
        fb->map_state = FALSE;
	fb->type_combo_glist = NULL;
	fb->accelgrp = accelgrp = gtk_accel_group_new();
	fb->last_dir_path = NULL;

	/* Create the file browser. */
	fb->file_browser = w = gtk_file_selection_new("File selection");
	fs = GTK_FILE_SELECTION(w);
        if(!GTK_WIDGET_REALIZED(w))
            gtk_widget_realize(w);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_TITLE | GDK_DECOR_MENU | GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
	}
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(FileBrowserDeleteEventCB), fb
        );

	/* Get parent to button widgets, destroy existing buttons
	 * and create our own.
	 */
	w = GTK_FILE_SELECTION(w)->ok_button;
	parent = (w != NULL) ? w->parent : NULL;
	if((parent != NULL) ? GTK_IS_BOX(parent) : FALSE)
	{
	    GtkBox *box = GTK_BOX(parent);

	    /* Modify box. */
	    gtk_box_set_homogeneous(box, FALSE);
	    gtk_box_set_spacing(box, border_major);

	    /* Create our own buttons. */
            /* Refresh button. */
            fb->refresh_btn = w = GUIButtonPixmapLabelH(
                (u_int8_t **)icon_reload_20x20_xpm, "Refresh",
                (gpointer *)&fb->refresh_btn_label
            );
            gtk_widget_set_usize(w, FB_BTN_WIDTH, FB_BTN_HEIGHT);
            GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
            gtk_box_pack_end(box, w, FALSE, FALSE, 0);
            gtk_signal_connect(
                GTK_OBJECT(w), "clicked",
                GTK_SIGNAL_FUNC(FileBrowserRefreshCB), fb
            );
            gtk_accel_group_add(
                accelgrp, GDK_F5, 0, GTK_ACCEL_VISIBLE,
                GTK_OBJECT(w), "clicked"
            );
            gtk_widget_show(w);

	    /* OK button. */
	    fb->ok_btn = w = GUIButtonPixmapLabelH(
		(u_int8_t **)icon_ok_20x20_xpm, "OK",
		(gpointer *)&fb->ok_btn_label
	    );
	    gtk_widget_set_usize(w, FB_BTN_WIDTH, FB_BTN_HEIGHT);
	    GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	    gtk_box_pack_end(box, w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(FileBrowserOKCB), fb
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    gtk_widget_show(w);

	    /* Cancel button. */
            fb->cancel_btn = w = GUIButtonPixmapLabelH(
                (u_int8_t **)icon_cancel_20x20_xpm,
#ifdef PROG_LANGUAGE_ENGLISH
                "Cancel",
#endif
#ifdef PROG_LANGUAGE_SPANISH
                "Cancela",
#endif
#ifdef PROG_LANGUAGE_FRENCH
                "Annuler",
#endif
#ifdef PROG_LANGUAGE_GERMAN
                "Auf",
#endif
#ifdef PROG_LANGUAGE_ITALIAN
                "Annulla",
#endif
#ifdef PROG_LANGUAGE_NORWEGIAN
                "Kanseller",
#endif
#ifdef PROG_LANGUAGE_PORTUGUESE
                "cancelamento",
#endif
		(gpointer *)&fb->cancel_btn_label
            );
            gtk_widget_set_usize(w, FB_BTN_WIDTH, FB_BTN_HEIGHT);
            GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
            gtk_box_pack_end(box, w, FALSE, FALSE, 0);
            gtk_signal_connect(
                GTK_OBJECT(w), "clicked",
                GTK_SIGNAL_FUNC(FileBrowserCancelCB), fb
            );
            gtk_accel_group_add(
                accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
                GTK_OBJECT(w), "clicked"
            );
	    gtk_widget_show(w);

	    /* Destroy existing buttons. */
	    w = fs->ok_button;
	    fs->ok_button = fb->ok_btn;
	    if(w != NULL)
		gtk_widget_destroy(w);

            w = fs->cancel_button;
	    fs->cancel_button = fb->cancel_btn;
            if(w != NULL)
                gtk_widget_destroy(w);
	}


	/* Create type pull down. */
	w = gtk_table_new(1, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(fs->main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent = w;

        w = gtk_label_new("Type:");
	gtk_table_attach(GTK_TABLE(parent), w,
	    0, 1,
	    0, 1,
	    0,
            0,
	    2, 2
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_widget_show(w);

	fb->type_combo = w = gtk_combo_new();
        gtk_table_attach(GTK_TABLE(parent), w,
            1, 2,
            0, 1,
            GTK_EXPAND | GTK_FILL,            
            0,
            2, 2
        );
	gtk_combo_disable_activate(GTK_COMBO(w));
	gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(w)->entry), FALSE);
	gtk_signal_connect(
	    GTK_OBJECT(GTK_COMBO(w)->entry), "activate",
	    GTK_SIGNAL_FUNC(FileBrowserTypeChangeCB),
	    fb
	);
        gtk_signal_connect(
            GTK_OBJECT(GTK_COMBO(w)->entry), "changed",
            GTK_SIGNAL_FUNC(FileBrowserTypeChangeCB),
            fb
        );
	gtk_widget_show(w);

	/* Set default type string. */
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(w)->entry), FB_DEFAULT_TYPE_STR);

	/* Create default glist on type combo. */
	glist = NULL;
	glist = g_list_append(glist, g_strdup(FB_DEFAULT_TYPE_STR));
	gtk_combo_set_popdown_strings(GTK_COMBO(w), glist);

	/* Update type combo's glist pointer. */
	fb->type_combo_glist = glist;
	glist = NULL;

	return(0);
}

/*
 *      Sets dialog to be a transient for the given toplevel window
 *      widget w. If w is NULL then no transient for will be unset.
 */
void FileBrowserSetTransientFor(GtkWidget *w)
{
	fb_data_struct *fb = &file_browser_data;

        if(!fb->initialized)
            return;

        if(fb->file_browser != NULL)
        {
            if(w != NULL)
            {
                /* Given widget if not NULL, must be a window. */
                if(!GTK_IS_WINDOW(GTK_OBJECT(w)))
                    return;

                gtk_window_set_modal(
                    GTK_WINDOW(fb->file_browser), TRUE
                );
                gtk_window_set_transient_for(
                    GTK_WINDOW(fb->file_browser), GTK_WINDOW(w)
                );
            }
            else
            {
                gtk_window_set_modal(
                    GTK_WINDOW(fb->file_browser), FALSE
                );
                gtk_window_set_transient_for(
                    GTK_WINDOW(fb->file_browser), NULL
                );
            }
        }
}

/*
 *      Returns TRUE if currently blocking for query.
 */
gbool FileBrowserIsQuery(void) 
{
        if(block_loop_level > 0)
            return(TRUE);
        else
            return(FALSE);
}

/*
 *      Ends query if any and returns a not available response.
 */
void FileBrowserBreakQuery(void)
{
        fb_got_user_response = FALSE;

        /* Break out of an additional blocking loops. */
        while(block_loop_level > 0)
        {
            gtk_main_quit();
            block_loop_level--;
        }
        block_loop_level = 0;
}

/*
 *	Maps the file browser and sets up the inital values.
 *
 *	Returns TRUE if a path was selected or FALSE if user canceled.
 *
 *	For most values that are set NULL, the value is left unchanged.
 *	All given values are coppied.
 *
 *	If type is set to NULL however, then the type list on the file
 *	browser will be left empty.
 *
 *	All returned pointer values should be considered statically
 *	allocated. The returned pointer for type_rtn may point to
 *	a structure in the input type list.
 */
gbool FileBrowserGetResponse(
	const gchar *title,
	const gchar *ok_label, const gchar *cancel_label,
	const gchar *path,
	fb_type_struct **type, gint total_types,
        gchar ***path_rtn, gint *path_total_rtns,
	fb_type_struct **type_rtn
)
{
	gint i;
	GtkWidget *w;
	GtkFileSelection *fs;
	fb_type_struct *t_ptr;
	fb_data_struct *fb = &file_browser_data;


        /* Do not handle response if already waiting for a response,
         * return with a not available response code.
         */
        if(block_loop_level > 0)
	{
	    if(path_rtn != NULL)
		(*path_rtn) = NULL;
	    if(path_total_rtns != NULL)
		(*path_total_rtns) = 0;
	    if(type_rtn != NULL)
		(*type_rtn) = NULL;

            return(FALSE);
	}

	/* Reset global responses values. */
	fb_got_user_response = FALSE;

	for(i = 0; i < fb_response_total_paths; i++)
	    g_free(fb_response_path[i]);
	g_free(fb_response_path);
	fb_response_path = NULL;
	fb_response_total_paths = 0;

	g_free(fb_response_type.ext);
        g_free(fb_response_type.name);
        memset(&fb_response_type, 0x00, sizeof(fb_type_struct));


	/* Reset returns. */
	if(path_rtn != NULL)
	    (*path_rtn) = NULL;
	if(path_total_rtns != NULL)
	    (*path_total_rtns) = 0;
	if(type_rtn != NULL)
	    (*type_rtn) = NULL;

	/* File browser must be initialized. */
	if(!fb->initialized)
	    return(fb_got_user_response);

	/* Get pointer to file browser widget. */
	w = fb->file_browser;
	if(w == NULL)
            return(fb_got_user_response);
	else
	    fs = GTK_FILE_SELECTION(w);


	/* Update title if specified. */
	if(title != NULL)
	    gtk_window_set_title(GTK_WINDOW(fs), title);

	/* Update ok button label if specified. */
	w = fb->ok_btn_label;
	if((ok_label != NULL) && (w != NULL))
	    gtk_label_set_text(GTK_LABEL(w), ok_label);

        /* Update cancel button label if specified. */
        w = fb->cancel_btn_label;
        if((cancel_label != NULL) && (w != NULL))
            gtk_label_set_text(GTK_LABEL(w), cancel_label);

	/* Set initial path if specified. */
	if(path != NULL)
	{
	    gboolean dir_changed = TRUE;
	    struct stat stat_buf;
	    gchar *s = g_strdup(path);

	    /* Check if path exists and if it's a dir. */
	    if(!stat(path, &stat_buf))
	    {
		if(S_ISDIR(stat_buf.st_mode))
		{
		    g_free(s);
		    s = g_strdup_printf("%s%c", path, DIR_DELIMINATOR);

		    /* Directories still the same? */
		    if((fb->last_dir_path != NULL) ?
			!strcmp(s, fb->last_dir_path) : FALSE
		    )
		    {
			dir_changed = FALSE;
		    }
		    else
		    {
			/* New dir different, so update last dir path. */
			g_free(fb->last_dir_path);
			fb->last_dir_path = g_strdup(s);
		    }
		}
	    }

	    /* Set new path only if directory changed. */
	    if(dir_changed)
		gtk_file_selection_set_filename(fs, s);

	    g_free(s);
	}

	/* Set first type to local global fb_response_type if
	 * available.
	 */
	if((type != NULL) && (total_types > 0))
	{
	    t_ptr = type[0];
	    if(t_ptr != NULL)
	    {
		/* Set that value to our local global fb_response_type. */
		fb_response_type.ext = ((t_ptr->ext == NULL) ?
		    NULL : g_strdup(t_ptr->ext)
		);
	    }

	    /* Since a new set of file extension types are given, we need
	     * need to clear and update the type combo box.
	     */
	    w = fb->type_combo;
	    if(w != NULL)
	    {
		gchar *buf;
		GList *glist = NULL;
		gchar first_text[(2 * NAME_MAX) + 80];

		*first_text = '\0';

		/* Iterate through all given file extension types. */
		for(i = 0; i < total_types; i++)
                {
                    t_ptr = type[i];
                    if(t_ptr == NULL)
                        continue;

		    if(t_ptr->ext == NULL)
			continue;

		    /* Format type string as buf. */
		    if((t_ptr->name != NULL) ?
			(*t_ptr->name != '\0') : FALSE
		    )
			buf = g_strdup_printf(
			    "%s (%s)", t_ptr->name, t_ptr->ext
			);
		    else
			buf = g_strdup(t_ptr->ext);

		    /* Add type string buf to the glist. */
		    glist = g_list_append(glist, buf);

		    /* Is this the first extension given? If so then
		     * record it in the first_text buffer.
		     */
		    if(i == 0)
		    {
			strncpy(first_text, buf, (2 * NAME_MAX) + 80);
			first_text[(2 * NAME_MAX) + 80 - 1] = '\0';
		    }
		}

		/* New glist valid? */
		if(glist != NULL)
		{
		    GList *old_glist;

		    /* Set new glist to type combo. */
		    gtk_combo_set_popdown_strings(GTK_COMBO(w), glist);

		    /* Delete old recorded glist and type combo current
		     * glist pointer to the new glist.
		     */
		    old_glist = fb->type_combo_glist;
		    fb->type_combo_glist = glist;
		    if(old_glist != NULL)
		    {
			g_list_foreach(old_glist, (GFunc)g_free, NULL);
			g_list_free(old_glist);
			old_glist = NULL;
		    }

		    /* Set first type text string to combo's entry. */
		    gtk_entry_set_text(
			GTK_ENTRY(GTK_COMBO(w)->entry),
			first_text
		    );
		}	/* New glist valid? */

	    }
	}


        /* Map file browser as needed. */
        FileBrowserMap();

        /* Block GUI untill response. */
	block_loop_level++;
        gtk_main();

        /* Unmap file browser just in case it was not unmapped from
	 * any of the callbacks.
	 */
        FileBrowserUnmap();

        /* Break out of an additional blocking loops. */
        while(block_loop_level > 0)
        {
            gtk_main_quit();
            block_loop_level--;
        }
        block_loop_level = 0;


	/* Begin setting returns. */

	/* Response path returns. */
	if(path_rtn != NULL)
	    *path_rtn = fb_response_path;
	if(path_total_rtns != NULL)
	    *path_total_rtns = fb_response_total_paths;

	/* Set up type return? */
	if(type_rtn != NULL)
	{
	    /* The member ext needs to be allocated, otherwise we
	     * need to report NULL as the file extension type return.
	     */
	    if(fb_response_type.ext == NULL)
	    {
		/* No extension available. */
		*type_rtn = NULL;
	    }
	    else
	    {
		/* Go through types list. */
		for(i = 0; i < total_types; i++)
		{
		    t_ptr = type[i];
		    if(t_ptr == NULL)
			continue;

		    if(t_ptr->ext == NULL)
			continue;

		    /* Extension types match? */
		    if(!strcasecmp(t_ptr->ext, fb_response_type.ext))
		    {
			/* Set matched values. */
			g_free(fb_response_type.name);
			fb_response_type.name = ((t_ptr->name == NULL) ?
			    NULL : g_strdup(t_ptr->name)
			);
		    }
		}

                *type_rtn = &fb_response_type;
	    }
	}


	return(fb_got_user_response);
}


/*
 *	Maps the file browser as needed.
 */
void FileBrowserMap(void)
{
	GtkWidget *w;
        fb_data_struct *fb = &file_browser_data;


        if(fb == NULL)
            return;

        if(!fb->initialized)
            return;

	w = fb->file_browser;
	gtk_widget_show_raise(w);
	fb->map_state = TRUE;
}

/*
 *	Unmaps the file browser as needed.
 */
void FileBrowserUnmap(void)
{
        fb_data_struct *fb = &file_browser_data;


	if(fb == NULL)
	    return;

	if(!fb->initialized)
	    return;

	if(fb->map_state)
	{
	    GtkWidget *w = fb->file_browser;

	    if(w != NULL)
		gtk_widget_hide(w);

	    fb->map_state = FALSE;
	}
}

/*
 *	Deallocates file browser resources.
 */
void FileBrowserShutdown(void)
{
	gint i;
	GtkWidget **w;
	GList *glist;
        fb_data_struct *fb = &file_browser_data;


	/* Reset local globals. */
        fb_got_user_response = FALSE; 

        for(i = 0; i < fb_response_total_paths; i++)
            g_free(fb_response_path[i]);
        g_free(fb_response_path);
        fb_response_path = NULL;
        fb_response_total_paths = 0;

        g_free(fb_response_type.ext);
        g_free(fb_response_type.name);
        memset(&fb_response_type, 0x00, sizeof(fb_type_struct));

        /* Break out of an additional blocking loops. */
        while(block_loop_level > 0)
        {
            gtk_main_quit();
            block_loop_level--;
        }
        block_loop_level = 0;


	/* Is file browser initialized? */
	if(fb->initialized)
	{
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Begin destroying widgets. */

	    /* File extension type combo. */
	    w = &fb->type_combo;
	    DO_DESTROY_WIDGET
	    /* File extension type combo's list. */
	    glist = fb->type_combo_glist;
	    fb->type_combo_glist = NULL;
	    if(glist != NULL)
	    {
		g_list_foreach(glist, (GFunc)g_free, NULL);
		g_list_free(glist);
	    }

	    /* Actual file browser. */
	    w = &fb->file_browser;
	    DO_DESTROY_WIDGET

	    if(fb->accelgrp != NULL)
	    {
		gtk_accel_group_unref(fb->accelgrp);
		fb->accelgrp = NULL;
	    }

	    g_free(fb->last_dir_path);
	    fb->last_dir_path = NULL;
#undef DO_DESTROY_WIDGET
	}

	/* Clear file browser structure. */
	memset(fb, 0x00, sizeof(fb_data_struct));
}


/*
 *	Convience function to allocate a new file browser file extension
 *	type structure and append it to the given list. The given list
 *	will be modified and the index number of the newly allocated
 *	structure will be returned.
 *
 *	Can return -1 on error.
 */
gint FileBrowserTypeListNew(
	fb_type_struct ***list, gint *total,
        const gchar *ext,	/* Space separated list of extensions. */
        const gchar *name	/* Descriptive name. */
)
{
	gint n;
	fb_type_struct *fb_type_ptr;


	if((list == NULL) || (total == NULL))
	    return(-1);

	/* Increase total. */
	if(*total < 0)
	    *total = 0;
	n = *total;
	*total = n + 1;

	/* Allocate more pointers. */
	*list = (fb_type_struct **)g_realloc(
	    *list,
	    (*total) * sizeof(fb_type_struct *)
	);
	if(*list == NULL)
	{
	    *total = 0;
	    return(-1);
	}

	/* Allocate new structure. */
	fb_type_ptr = (fb_type_struct *)g_malloc0(sizeof(fb_type_struct));
	(*list)[n] = fb_type_ptr;
	if(fb_type_ptr == NULL)
	{
	    *total = n;
	    return(-1);
	}

	/* Set values. */
	if(ext != NULL)
	    fb_type_ptr->ext = g_strdup(ext);
	if(name != NULL)
	    fb_type_ptr->name = g_strdup(name);

	return(n);
}

/*
 *	Deletes a dynamically allocated file browser file extensions
 *	type list.
 */
void FileBrowserDeleteTypeList(
        fb_type_struct **t, gint total 
)
{
	gint i;
	fb_type_struct *t_ptr;

	for(i = 0; i < total; i++)
	{
	    t_ptr = t[i];
	    if(t_ptr == NULL)
		continue;

	    g_free(t_ptr->ext);
	    g_free(t_ptr->name);
	    g_free(t_ptr);
	}
	g_free(t);
}
